<?php
/**
 * @package     CSVI
 * @subpackage  VirtueMart
 *
 * @author      RolandD Cyber Produksi <contact@rolandd.com>
 * @copyright   Copyright (C) 2006 - @year@ RolandD Cyber Produksi. All rights reserved.
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 * @link        https://rolandd.com
 */

namespace phocacart\com_phocacart\model\import;

use Com_phocacartHelperCategory;
use Com_PhocacartHelperCom_Phocacart;
use CsviHelperImportFields;
use Exception;
use Joomla\CMS\Filter\OutputFilter;
use Joomla\CMS\Language\Text;
use PhocacartTableManufacturer;
use PhocacartTableProduct;
use PhocacartTableProductImage;
use PhocacartTableTax;
use Rocloud\Common\File;
use Rocloud\Common\Image;

defined('_JEXEC') or die;

/**
 * Main processor for handling product details.
 *
 * @package     CSVI
 * @subpackage  Phocacart
 * @since       8.0.0
 */
class Product extends \RantaiImportEngine
{
	/**
	 * The fields helper
	 *
	 * @var    CsviHelperImportFields
	 * @since  6.0
	 */
	protected $fields;

	/**
	 * Phoca Cart helper
	 *
	 * @var    Com_PhocacartHelperCom_Phocacart
	 * @since  6.0
	 */
	protected $helper = [];

	/**
	 * Collect deleted product ids
	 *
	 * @var    array
	 *
	 * @since  7.7.0
	 */
	protected $deletedProductIds = [];

	/**
	 * The product table
	 *
	 * @var    PhocacartTableProduct
	 * @since  6.0
	 */
	private $productTable;

	/**
	 * The product images table
	 *
	 * @var    PhocacartTableProductImage
	 * @since  6.0
	 */
	private $productImageTable;

	/**
	 * The manufacturer table
	 *
	 * @var    PhocacartTableManufacturer
	 * @since  6.0
	 */
	private $manufacturerTable;

	/**
	 * The tax table
	 *
	 * @var    PhocacartTableTax
	 * @since  6.0
	 */
	private $taxTable;

	/**
	 * Category model
	 *
	 * @var    Com_phocacartHelperCategory
	 * @since  8.0.0
	 */
	private $categoryModel;

	/**
	 * Here starts the processing.
	 *
	 * @return  boolean  Returns true on success | False on failure.
	 *
	 * @since   3.0
	 * @throws  Exception
	 *
	 */
	public function getStart()
	{
		if (!$this->fields->getProcessRecord())
		{
			$this->loaded = false;

			return true;
		}

		foreach ($this->fields->getData() as $fields)
		{
			foreach ($fields as $name => $details)
			{
				$value = $details->value;

				switch ($name)
				{
					case 'date':
					case 'date_update':
						$this->setState($name, $this->convertDate($value));
						break;
					case 'price':
					case 'price_original':
						$this->setState($name, $this->toPeriod($value));
						break;
					case 'category_id':
					case 'category_path':
						if (strlen(trim($value)) > 0)
						{
							if (stripos($value, '|') > 0)
							{
								$category_ids[$name] = explode('|', $value);
							}
							else
							{
								$category_ids[$name][] = $value;
							}

							$this->setState('category_ids', $category_ids);
						}

						$this->setState($name, $value);
						break;
					case 'price_with_tax':
						$this->setState($name, $this->cleanPrice($value));
						break;
					case 'published':
						switch ($value)
						{
							case 'n':
							case 'no':
							case 'N':
							case 'NO':
							case '0':
								$value = 0;
								break;
							default:
								$value = 1;
								break;
						}

						$this->setState($name, $value);
						break;
					default:
						$this->setState($name, $value);
						break;
				}
			}
		}

		$field = $this->template->get('update_based_on', 'sku');

		if ($this->getState($field, false))
		{
			$this->productTable->setKeyName($field);
		}
		else
		{
			$field = 'id';
		}

		if ($this->getState($field, false))
		{
			// Set the record identifier
			$this->recordIdentity = $this->getState($field, '');

			// Load the current product data
			$this->loaded = true;

			if ($this->productTable->load($this->getState($field, false)))
			{
				if (!$this->template->get('overwrite_existing_data'))
				{
					$this->log->addStats(
						'skipped', Text::sprintf('COM_CSVI_DATA_EXISTS_PRODUCT_SKU', $this->getState($field, ''))
					);
					$this->loaded = false;
				}
				else
				{
					$this->setState('id', $this->productTable->get('id', 0));
				}
			}
			else
			{
				$productId = $this->getState('id', false);

				if (!$this->getState($field, false) && empty($productId))
				{
					$this->log->addStats('incorrect', 'COM_CSVI_DEBUG_NO_SKU');
					$this->log->add('COM_CSVI_DEBUG_NO_SKU');

					$this->loaded = false;
				}
				else
				{
					// Product is not found so we need to reset to the primary key field
					$this->productTable->setKeyName('id');
					$this->log->add(Text::sprintf('COM_CSVI_DEBUG_PROCESS_SKU', $this->recordIdentity), false);
				}
			}
		}
		else
		{
			$this->loaded = false;

			$this->log->addStats(
				'skipped',
				Text::sprintf(
					'COM_CSVI_MISSING_REQUIRED_FIELDS_PRODUCT', $this->template->get('update_based_on', 'sku')
				),
				'product'
			);
		}

		return true;
	}

	/**
	 * Process each record and store it in the database.
	 *
	 * @return  bool  Returns true if all is OK | Returns false if no product SKU or product ID can be found.
	 *
	 * @since   3.0
	 */
	public function getProcessRecord()
	{
		if ($this->loaded)
		{
			// Get the needed data
			$productId      = $this->getState('id', false);
			$productDelete  = $this->getState('product_delete', 'N');
			$categoryDelete = $this->getState('category_delete', 'N');

			// If user wants to remove category for a product
			if ($productId && $categoryDelete === 'Y')
			{
				$this->deleteProductCategory();
			}

			// User wants to delete the product
			if ($productId && $productDelete === 'Y')
			{
				$this->deleteProduct();
			}
			elseif (!$productId && $productDelete === 'Y')
			{
				$this->log->addStats(
					'skipped', Text::sprintf('COM_CSVI_NO_PRODUCT_ID_NO_DELETE', $this->recordIdentity)
				);
			}
			elseif (!$productId && $this->template->get('ignore_non_exist'))
			{
				// Do nothing for new products when user chooses to ignore new products
				$this->log->addStats(
					'skipped', Text::sprintf('COM_CSVI_DATA_EXISTS_IGNORE_NEW', $this->recordIdentity)
				);
			}
			else
			{
				// Process discount
//				if ($this->getState('product_discount', false))
//				{
//					$this->processDiscount();
//				}

				// Process tax
				$this->log->add('Product tax: ' . $this->getState('product_tax'), false);

				if ($this->getState('product_tax'))
				{
					$this->processTax();
				}

				// Process the ICEcat product features
				$this->setState('features', $this->fields->get('features'));

				if ($this->template->get('use_icecat', false, 'bool') && $this->getState('features', false))
				{
					$this->icecatFeatures();
				}

				$this->manufacturerImport();

				// Process product info
				if ($this->productQuery())
				{
					// Handle the shopper group(s)
//					$this->processShopperGroup();

					// Handle the images
					$this->processMedia();

					// Check if the price is to be updated
//					if ($this->getState('product_price', false) !== false
//						|| $this->getState('price_with_tax', false) !== false
//						|| $this->getState('product_override_price', false) !== false
//					)
//					{
//						$this->priceQuery();
//					}

					// Delete related products if user wants to
//					if ($this->template->get('delete_related_products', false))
//					{
//						$this->productCustomfieldTable->deleteRelated($this->virtuemart_product_id, $this->virtuemart_vendor_id, $this->helper->getRelatedId('R'));
//						$this->log->add('Removed existing related products for product' . $this->virtuemart_product_id);
//					}

					/**
					 * Process related products/categories
					 * Related products are first input in the database as SKU
					 * At the end of the import, this is converted to product ID
					 */

					if ($this->getState('related_products'))
					{
						$this->processRelatedProducts();
					}

					// Process category path
					if (($this->getState('category_path', false) || $this->getState('category_id', false))
						&& $this->getState('category_ids', false))
					{
						if (null === $this->categoryModel)
						{
							$this->categoryModel = new Com_phocacartHelperCategory(
								$this->db,
								$this->template,
								$this->log,
								$this->csvihelper,
								$this->fields,
								$this->helper,
								$this->helperconfig,
								$this->userId
							);
						}

						$this->categoryModel->getStart();

						// Check the categories
						// Do we have IDs
						if (array_key_exists('category_id', $this->getState('category_ids')))
						{
							$this->categoryModel->checkCategoryPath(
								$this->getState('id'),
								false,
								$this->category_ids['category_id'],
								$this->getState('product_ordering')
							);
						}
						elseif (array_key_exists('category_path', $this->getState('category_ids')))
						{
							$this->categoryModel->checkCategoryPath(
								$this->getState('id'),
								$this->category_ids['category_path'],
								false,
								$this->getState('ordering')
							);
						}
					}
				}
			}

			return true;
		}
		else
		{
			return false;
		}
	}

	/**
	 * Delete a product category reference.
	 *
	 * @return  void.
	 *
	 * @since   8.0.0
	 */
	private function deleteProductCategory()
	{
		// Get the product ID
		$productId = $this->getState('id');

		// Delete product category reference
		$query = $this->db->getQuery(true)
			->delete($this->db->quotename('#__phocacart_product_categories'))
			->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
		$this->db->setQuery($query)->execute();
		$this->log->add('Remove product category references');
	}

	/**
	 * Delete a product and its references.
	 *
	 * @return  bool  Always returns true.
	 *
	 * @since   4.0
	 */
	private function deleteProduct()
	{
		// Deletion based on the real primary key
		$this->productTable->setKeyName('id');

		// Get the product ID
		$productId = $this->getState('id');

		// Delete the product
		if ($this->productTable->delete($productId))
		{
			$query = $this->db->getQuery(true);
			$query->delete($this->db->quotename('#__phocacart_product_categories'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete product category references');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_images'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete product image references');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_files'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete product file references');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_featured'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete the product featured status');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_discounts'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete the product discounts');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_point_groups'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete the product point groups');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_price_groups'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete the product price groups');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_price_history'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete the product price history');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_related'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete the product related');

			$query->clear();
			$query->delete($this->db->quotename('#__phocacart_product_stock'));
			$query->where($this->db->quotename('product_id') . ' = ' . (int) $productId);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete the product stock');

			$this->log->addStats(
				'deleted', Text::sprintf('COM_CSVI_PRODUCT_DELETED', $this->getState('recordIdentity'))
			);
		}
		else
		{
			$this->log->addStats(
				'incorrect', Text::sprintf('COM_CSVI_PRODUCT_NOT_DELETED', $this->getState('recordIdentity'))
			);
		}

		return true;
	}

	/**
	 * Stores a product discount.
	 *
	 * @return  bool Returns true on OK | False on failure.
	 *
	 * @since   3.0
	 */
	private function processDiscount()
	{
		$this->log->add('COM_CSVI_DEBUG_PROCESSING_DISCOUNT');

		// Clear the calcs from any data
		$this->calcsTable->reset();

		// Determine if the discount field is a percentage
		if ($this->getState('product_discount', false))
		{
			if (substr($this->getState('product_discount'), -1, 1) === '%')
			{
				$this->calcsTable->calc_value_mathop = '-%';
				$this->calcsTable->calc_value        = substr($this->toPeriod($this->product_discount), 0, -1);
			}
			else
			{
				$this->calcsTable->calc_value_mathop = '-';
				$this->calcsTable->calc_value        = $this->cleanPrice($this->product_discount);
			}
		}

		if ((null !== $this->calcsTable->calc_value) && $this->calcsTable->calc_value > 0)
		{
			// Add the discount fields
			$this->calcsTable->publish_up   = $this->getState('product_discount_date_start');
			$this->calcsTable->publish_down = $this->getState('product_discount_date_end');

			// Add a description to the discount
			$this->calcsTable->calc_name              = $this->getState('product_discount');
			$this->calcsTable->calc_descr             = $this->getState('product_discount');
			$this->calcsTable->calc_shopper_published = 1;
			$this->calcsTable->calc_vendor_published  = 1;
			$this->calcsTable->calc_currency          = $this->productPriceTable->product_currency;

			if (!$this->getState('calc_kind'))
			{
				$this->calcsTable->calc_kind = 'DBTax';
			}
			else
			{
				$this->calcsTable->calc_kind = $this->getState('calc_kind');
			}

			// Check if a discount already exists
			$this->calcsTable->check();

			// Store the discount
			if (!$this->calcsTable->store())
			{
				$this->log->add('COM_CSVI_DEBUG_ADD_DISCOUNT', true);

				return false;
			}

			$this->log->add('COM_CSVI_DEBUG_ADD_DISCOUNT', true);

			// Fill the product information with the discount ID
			$this->setState('product_discount_id', $this->calcsTable->virtuemart_calc_id);
		}
		else
		{
			$this->log->add('COM_CSVI_DEBUG_NO_DISCOUNT');
		}

		return true;
	}

	/**
	 * Process a tax rate.
	 *
	 * @return  void.
	 *
	 * @since   8.0.0
	 */
	private function processTax()
	{
		if ($this->getState('product_tax'))
		{
			$this->log->add('Processing product tax', false);

			$this->taxTable->reset();

			$this->taxTable->code             = $this->getState('product_tax');
			$this->taxTable->calculation_type = $this->getState('calculation_type', '1');

			if (!$this->taxTable->check())
			{
				$this->log->add('Adding a CSVI generated tax rule', false);

				$this->taxTable->title     = Text::_('COM_CSVI_AUTO_TAX_RATE');
				$this->taxTable->alias     = OutputFilter::stringURLUnicodeSlug($this->taxTable->title);
				$this->taxTable->published = 1;
				$this->taxTable->store();
			}

			$this->setState('tax_id', $this->calcsTable->id);
		}
	}

	/**
	 * Combine the ICEcat features.
	 *
	 * @return  void.
	 *
	 * @since   4.3
	 */
	private function icecatFeatures()
	{
		$table = '<table id="prod_features">';
		$table .= '<thead></thead>';
		$table .= '<tfoot></tfoot>';
		$table .= '<tbody>';

		foreach ($this->getState('features') as $details)
		{
			foreach ($details as $feature => $values)
			{
				$table .= '<tr><td colspan="2" class="feature">' . $feature . '</td></tr>';

				foreach ($values as $name => $value)
				{
					$table .= '<tr><td>' . $name . '</td><td>' . $value . '</td></tr>';
				}
			}
		}

		$table .= '</tbody>';
		$table .= '</table>';

		// Add the table to the product desc
		$this->setState('product_desc', $this->getState('product_desc') . $table);
	}

	/**
	 * Creates either an update or insert SQL query for a product.
	 *
	 * @return  bool true if the query executed successful|false if the query failed.
	 *
	 * @since   3.0
	 *
	 * @throws  Exception
	 */
	private function productQuery()
	{
		if ($this->getState('stock', false))
		{
			// Split the modification
			$operation = substr($this->getState('stock'), 0, 1);
			$value     = substr($this->getState('stock'), 1);

			// Get the database value
			$stock = $this->productTable->stock;

			// Check what modification we need to do and apply it
			switch ($operation)
			{
				case '+':
					$stock += $value;
					break;
				case '-':
					$stock -= $value;
					break;
				case '/':
					$stock /= $value;
					break;
				case '*':
					$stock *= $value;
					break;
				default:
					// Assign the current price to prevent it being overwritten
					$stock = $this->getState('stock');
					break;
			}

			$this->setState('stock', $stock);
		}

		if (!$this->getState('alias', false) && $this->getState('title', false))
		{
			$this->productTable->checkAlias();
			$this->getState('alias', $this->productTable->alias);
		}

		$this->productTable->language = $this->template->get('language', '*');

		// Bind the initial data
		$this->productTable->bind($this->state);

		// Set the modified date as we are modifying the product
		if (!$this->getState('modified_on'))
		{
			$this->productTable->modified    = $this->getState('modified', $this->date->toSql());
			$this->productTable->modified_by = $this->getState('modified_by', $this->userId);
		}

		// Add a creating date if there is no product_id
		if (!$this->getState('id'))
		{
			$this->productTable->created    = $this->getState('created', $this->date->toSql());
			$this->productTable->created_by = $this->getState('created_by', $this->userId);

			// Process default values
			$defaults = ['published'];

			foreach ($defaults as $field)
			{
				switch ($field)
				{
					case 'published':
						if (!$this->getState($field, false))
						{
							$this->productTable->$field = 0;
						}
						break;
				}
			}
		}

		try
		{
			// Store the product
			$this->productTable->store(true);
		}
		catch (Exception $e)
		{
			$this->log->addStats('error', Text::_('COM_CSVI_CANNOT_ADD_PRODUCTS'));
			$this->log->add($e->getMessage(), false);

			// Check if error is for duplicate sku and show a friendly message
			if (strpos($e->getMessage(), 'Duplicate entry') !== false)
			{
				$this->log->addStats('error', Text::_('COM_CSVI_DUPLICATE_SKU_PRODUCTS'));
			}

			return false;
		}

		// Set the product ID
		$this->setState('id', $this->productTable->get('id'));

		return true;
	}

	/**
	 * Process the shopper groups.
	 *
	 * @return  bool Returns true on OK | False on failure.
	 *
	 * @since   4.5.2
	 */
	private function processShopperGroup()
	{
		// Check if user wants to remove any assigned shopper groups before importing
		if ($this->template->get('delete_shopper_groups'))
		{
			if (!$this->getState('virtuemart_product_id', false))
			{
				return false;
			}

			$query = $this->db->getQuery(true)
				->delete($this->db->quotename('#__virtuemart_product_shoppergroups'))
				->where(
					$this->db->quotename('virtuemart_product_id') . ' = ' . (int) $this->getState(
						'virtuemart_product_id', false
					)
				);
			$this->db->setQuery($query)->execute();
			$this->log->add('Delete existing product shopper groups to insert new ones');
			$this->log->addStats('information', 'COM_CSVI_DELETE_PRODUCT_SHOPPER_GROUPS');

			// Update has_shoppergroups value to 0
			$this->setState('has_shoppergroups', 0);
		}

		if ($this->getState('shopper_group_name', false))
		{
			// Get the shopper group names
			$names = explode('|', $this->getState('shopper_group_name'));

			foreach ($names as $name)
			{
				$data                               = array();
				$data['virtuemart_shoppergroup_id'] = $this->helper->getCustomerGroupId($name);
				$data['virtuemart_product_id']      = $this->getState('virtuemart_product_id');

				// Process only if shopper group id is not zero
				if ($data['virtuemart_shoppergroup_id'])
				{
					// Set the shopper group ID for other updates
					$this->setState('virtuemart_shoppergroup_id', $data['virtuemart_shoppergroup_id']);

					// Bind the data to check
					$this->productShoppergroupTable->bind($data);

					// Check if a product - shopper group relation exists
					if (!$this->productShoppergroupTable->check())
					{
						if (!$this->productShoppergroupTable->store())
						{
							$this->log->addStats(
								'incorrect',
								Text::sprintf(
									'COM_CSVI_PRODUCT_SHOPPERGROUP_NOT_ADDED',
									$this->productShoppergroupTable->getError()
								)
							);

							return false;
						}
					}

					$this->setState('has_shoppergroups', 1);

					$this->productShoppergroupTable->reset();
				}
				else
				{
					$this->log->add('Shopper group name not found ' . $name, false);
				}
			}
		}

		return true;
	}

	/**
	 * Process media files.
	 *
	 * @return  void
	 *
	 * @since   4.0
	 *
	 * @throws  Exception
	 */
	private function processMedia(): void
	{
		$generateImage = $this->template->get('auto_generate_image_name', false);
		$fileUrl       = $this->getState('image', '');
		$productId     = $this->getState('id');

		if ($fileUrl || $generateImage)
		{
			if (('' === $fileUrl) && $generateImage)
			{
				$fileUrl = $this->createImageName();
			}

			$images     = explode('|', $fileUrl);
			$order      = explode('|', $this->getState('image_ordering'));
			$ordering   = 1;
			$imagePath = JPATH_BASE . '/' . $this->template->get(
					'file_location_product_files', $this->helperconfig->get('product_image_path', 'images/phocacartproducts')
				);

			// This block has been disabled as Phoca Cart does not yet support custom named thumbnails
			if (0)
			{
				$thumbnails = explode('|', $this->getState('image_thumb'));
			}

			if ($this->template->get('delete_product_images', false))
			{
				$this->deleteProductImages($productId);
			}

			$imageClass = new Image($this->log);
			$fileClass  = new File($this->log);

			foreach ($images as $key => $image)
			{
				try
				{
					$this->productImageTable->reset();
					$image = trim($image);

					if (empty($image))
					{
						continue;
					}

					$outputImage = $image;
					$imgCounter  = count($images) === 1 ? 0 : $key + 1;

					if ($generateImage)
					{
						$outputImage = $this->createImageName($imgCounter);
					}

					$fileClass
						->setBasePath(JPATH_BASE)
						->setDestinationPath($imagePath)
						->setSourceFileName($image)
						->setUserName($this->template->get('image_username'))
						->setPassword($this->template->get('image_password'))
						->setMethod($this->template->get('image_method', 'post'))
						->setCredentialType($this->template->get('image_credential_type', 'htaccess'))
						->setEncodeUrl($this->template->get('image_encodeurl', true))
						->setSaveImageOnServer($this->template->get('save_images_on_server', false))
						->setReDownload($this->template->get('redownload_external_image', true))
						->setVerifySsl($this->template->get('image_verify_ssl', true));

					if (!$generateImage && $fileClass->isRemote())
					{
						$outputImage = basename($image);
					}

					if ($this->template->get('update_only_media_link', false))
					{
						try
						{
							$setOrder = (array_key_exists($key, $order) && !empty($order[$key])) ? $order[$key] : $ordering;
							$this->updateMediaLink($imagePath . '/' . $image, $setOrder);
							$ordering++;
						}
						catch (Exception $exception)
						{
							$this->log->add($exception->getMessage(), false);
						}

						continue;
					}

					if ($this->template->get('process_image', false))
					{
						$fileClass
							->setDestinationFileName(basename($outputImage))
							->copyFile();

						$imageClass
							->setSourceFileName($fileClass->getDestinationFile())
							->setDestinationPath($fileClass->getDestinationPath())
							->setDestinationFileName($fileClass->getDestinationFileName())
							->setDestinationExtension($fileClass->getDestinationExtension());

						$fileClass->verifyImageExtension(
							$imageClass->getMimeType()
						);

						if ($this->template->get('convert_type', 'none') !== 'none')
						{
							$imageClass->setSourceFileName($fileClass->getDestinationFile());
							$imageClass->setDestinationExtension($this->template->get('convert_type'));
							$imageClass->convertImage();
							$fileClass->setDestinationExtension($this->template->get('convert_type'));
						}

						if ($this->template->get('full_resize', 0))
						{
							$imageClass
								->setOutputWidth($this->template->get('full_width'))
								->setOutputHeight($this->template->get('full_height'))
								->resize();
						}

						$fileClass->setCase($this->template->get('change_case', 'none'));
						$imageClass->setSourceFileName($fileClass->getDestinationFile());

						if ($this->template->get('full_watermark', '0') && $imageClass->isImage())
						{
							$imageClass
								->setSourceFileName($fileClass->getDestinationFile())
								->setWatermarkImage(JPATH_SITE . '/' . $this->template->get('full_watermark_image'))
								->setMarginRight($this->template->get('full_watermark_right', 0))
								->setMarginLeft($this->template->get('full_watermark_bottom', 0))
								->addWatermark();
						}

						if ($imageClass->isImage())
						{
							// This block has been disabled as Phoca Cart does not yet support custom named thumbnails
							if (0)
							{
								$thumbnail = $thumbnails[$key] ?? null;

								if ($thumbnail)
								{
									$dirname = dirname($thumbnail);

									if (strpos($imagePath, $dirname) !== false)
									{
										$imageLeftPath = str_replace($imagePath, '', $dirname . '/');

										if ($imageLeftPath)
										{
											$thumbnail = $imageLeftPath . basename($thumbnail);
										}
									}
								}

								if (empty($thumbnail) || $generateImage)
								{
									$thumbnail = $imagePath . '/thumbs/phoca_thumb_l_'
										. $fileClass->getDestinationFileName()
										. '.' . $fileClass->getDestinationExtension();
								}
							}

							$this->createThumbnail(
								$fileClass,
								$imagePath,
								$fileClass->getDestinationFileName() . '.' . $fileClass->getDestinationExtension()
							);
						}
					}

					if ($key === 0)
					{
						$this->productTable
							->set(
								'image', $fileClass->getDestinationFileName() . '.' . $fileClass->getDestinationExtension()
							);
						$this->productTable->store();
					}

					if ($key > 0)
					{
						$this->productImageTable->reset();
						$this->productImageTable->set('product_id', $productId);
						$this->productImageTable->set(
							'image', $fileClass->getDestinationFileName() . '.' . $fileClass->getDestinationExtension()
						);
						$this->productImageTable->set('ordering', $order[$key] ?? 0);
						$this->productImageTable->store();
					}

					$this->log->addStats('added', Text::_('COM_CSVI_PRODUCT_IMAGE_ADDED'));
				}
				catch (Exception $exception)
				{
					$this->log->add($exception->getMessage(), false);
					$this->log->addStats(
						'incorrect',
						$exception->getMessage()
					);
				}
			}
		}
	}

	/**
	 * Create the small, medium and large thumbnail images.
	 *
	 * Phoca has no options for creating thumbnails, they are always created in 3 formats:
	 *
	 * - Small (s)
	 * - Medium (m)
	 * - Large (l)
	 *
	 * The name is always prefixed by phoca_thumb_[s/m/l]_
	 *
	 * @param   File    $mainImage   The main image to create a thumbnail from
	 * @param   string  $outputPath  The path to store the images
	 * @param   string  $thumbnail   The name of the thumbnail
	 *
	 * @return  void
	 *
	 * @since   8.0.0
	 */
	private function createThumbnail(File $mainImage, string $outputPath, string $thumbnail)
	{
		$this->log->add('Create thumbnail', false);

		if (!$this->template->get('thumb_create', false))
		{
			return;
		}

		if (!$mainImage->destinationFileExistsLocal())
		{
			return;
		}

		$fileClass  = new File($this->log);
		$imageClass = new Image($this->log);
		$maxWidth   = $this->template->get('resize_max_width', 1024);
		$maxHeight  = $this->template->get('resize_max_height', 768);

		foreach (['s', 'm', 'l'] as $size)
		{
			$fileClass
				->setBasePath(JPATH_BASE)
				->setDestinationFileName('phoca_thumb_' . $size . '_' . $thumbnail)
				->setDestinationPath($outputPath . '/thumbs')
				->deleteDestinationFile()
				->checkDestinationFolderExists();

			if ($this->template->get('thumb_check_filetype', 0)
				&& $fileClass->getDestinationExtension() !== $mainImage->getDestinationExtension())
			{
				$fileClass->setDestinationExtension($mainImage->getDestinationExtension());
			}

			$fullImageSizes = getimagesize($mainImage->getDestinationFile());

			if ($thumbnail && ($fullImageSizes[0] < $maxWidth || $fullImageSizes[1] < $maxHeight))
			{
				$width  = 0;
				$height = 0;

				switch ($size)
				{
					case 's':
						$width  = $this->helperconfig->get('small_image_width');
						$height = $this->helperconfig->get('small_image_height');
						break;
					case 'm':
						$width  = $this->helperconfig->get('medium_image_width');
						$height = $this->helperconfig->get('medium_image_height');
						break;
					case 'l':
						$width  = $this->helperconfig->get('large_image_width');
						$height = $this->helperconfig->get('large_image_height');
						break;
				}

				if ($width > 0 && $height > 0)
				{
					$this->log->add(
						'Create thumbnail from ' . $mainImage->getDestinationFile() . ' to '
						. $fileClass->getDestinationFile(), false
					);
					$this->log->add('Create thumbnail for sizes ' . $width . 'x' . $height, false);
					$imageClass
						->setSourceFileName($mainImage->getDestinationFile())
						->setDestinationPath($fileClass->getDestinationPath())
						->setDestinationFileName($fileClass->getDestinationFileName())
						->setDestinationExtension($fileClass->getDestinationExtension())
						->setOutputWidth($width)
						->setOutputHeight($height)
						->findMimeType()
						->resize();
				}
			}
			else
			{
				$this->log->addStats(
					'incorrect', Text::sprintf(
					'COM_CSVI_THUMB_TOO_BIG', $maxWidth, $maxHeight, $fullImageSizes[0],
					$fullImageSizes[1]
				)
				);
				$this->log->add('Thumbnail is bigger than maximums set', false);
			}
		}
	}

	/**
	 * Delete existing product images for a given product.
	 *
	 * @param   int  $productId  The product ID to delete the images for
	 *
	 * @return  void
	 *
	 * @since   8.0.0
	 */
	private function deleteProductImages(int $productId): void
	{
		$query = $this->db->getQuery(true)
			->delete($this->db->quoteName('#__phocacart_product_images'))
			->where(
				$this->db->quoteName('product_id') . '=' . $productId
			);
		$this->db->setQuery($query)->execute();
		$this->log->add('Delete images');

		$query->clear()
			->update($this->db->quoteName('#__phocacart_products'))
			->set($this->db->quoteName('image') . ' = ' . $this->db->quote(''))
			->where(
				$this->db->quoteName('id') . '=' . $productId
			);
		$this->db->setQuery($query)->execute();
		$this->log->add('Delete images');
	}

	/**
	 * Update the media link for a product.
	 *
	 * @param   string  $image     The image
	 * @param   int     $ordering  The current ordering
	 *
	 * @return  void
	 *
	 * @since   8.0.0
	 */
	private function updateMediaLink(string $image, int $ordering): void
	{
		$this->productImageTable->image = (empty($outputPath))
			? basename($image)
			: $outputPath . basename($image);

		if (!$this->productImageTable->check())
		{
			throw new \RuntimeException('Image ' . $this->productImageTable->image . ' does not exist in media table');
		}

		$this->productImageTable->set('product_id', $this->getState('id'));
		$this->productImageTable->set('ordering', $ordering);
		$this->productImageTable->store();

		if ($this->productImageTable->getError())
		{
			throw new \RuntimeException($this->productImageTable->getError());
		}

		$this->log->add('Store product image relation', false);
	}

	/**
	 * Create image name.
	 *
	 * Check if the user wants to have CSVI VirtueMart create the image names if so
	 * create the image names without path.
	 *
	 * @param   int  $ordering  The number to apply to a generated image name.
	 *
	 * @return  string  The name of the image.
	 *
	 * @since   3.0
	 */
	private function createImageName(int $ordering = 0): string
	{
		$this->log->add('Generate an image name', false);

		// Create extension
		$ext = $this->template->get('autogenerateext');

		// Check if the user wants to convert the images to a different type
		switch ($this->template->get('type_generate_image_name'))
		{
			case 'product_sku':
				$this->log->add('Create name from product SKU', false);

				if ($this->getState('sku', false))
				{
					$name = $this->getState('sku');
				}
				else
				{
					$this->log->addStats('error', 'COM_CSVI_CANNOT_FIND_PRODUCT_SKU');

					return false;
				}
				break;
			case 'product_name':
				$this->log->add('Create name from product title', false);

				if ($this->productTable->get('title', false))
				{
					$name = str_replace(' ', '_', $this->productTable->get('title'));
				}
				else
				{
					$this->log->addStats('error', 'COM_CSVI_CANNOT_FIND_PRODUCT_NAME');

					return false;
				}
				break;
			case 'product_id':
				$this->log->add('Create name from product ID', false);

				if ($this->getState('id'))
				{
					$name = $this->getState('id');
				}
				else
				{
					$this->log->addStats('error', 'COM_CSVI_CANNOT_FIND_PRODUCT_ID');

					return false;
				}
				break;
			case 'random':
				$this->log->add('Create a random name', false);
				$name = mt_rand();
				break;
			default:
				$this->log->addStats('error', 'COM_CSVI_CANNOT_FIND_PRODUCT_SKU');

				return false;
				break;
		}

		// Build the new name
		if ($ordering > 0)
		{
			$imageName = $name . '_' . $ordering . '.' . $ext;
		}
		else
		{
			$imageName = $name . '.' . $ext;
		}

		$this->log->add('Created image name: ' . $imageName, false);

		// Check if the user is supplying image data
		if (!$this->getState('file_url', false))
		{
			$this->setState('file_url', $imageName);
		}

		return $imageName;
	}

	/**
	 * Creates either an update or insert SQL query for a product price.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 *
	 * @throws  Exception
	 */
	private function priceQuery()
	{
		// Check if we have a child product with an empty price (will use parents price)
		if ($this->child_product && ($this->product_price == 0) && (null === $this->price_with_tax)
			&& (null === $this->product_tax))
		{
			$this->log->add('COM_CSVI_DEBUG_CHILD_NO_PRICE', false);
		}
		else
		{
			// Remove existing product prices if user wants to
			if ($this->template->get('delete_existing_prices'))
			{
				$virtuemartProductId = $this->getState('virtuemart_product_id', false);

				if ($virtuemartProductId
					&& !in_array($virtuemartProductId, $this->deletedProductIds)
				)
				{
					$query = $this->db->getQuery(true)
						->delete($this->db->quotename('#__virtuemart_product_prices'))
						->where(
							$this->db->quotename('virtuemart_product_id') . ' = ' . (int) $this->getState(
								'virtuemart_product_id', false
							)
						);
					$this->db->setQuery($query)->execute();
					$this->log->add('Delete existing product prices to insert new ones');
					$this->deletedProductIds[] = $this->getState('virtuemart_product_id', false);

					// Update has_prices field to 0
					$this->setState('has_prices', 0);
				}
			}

			// Check if we have an override price, this is always excluding tax
			if ($this->getState('product_override_price', false) && ($this->getState('override', false) === false))
			{
				$this->setState('override', 1);
			}

			// Check if the price is including or excluding tax
			if (($this->getState('product_tax_id', false) || $this->getState('product_tax', false))
				&& $this->getState('price_with_tax', false)
				&& !$this->getState('product_price', false))
			{
				if (strlen($this->getState('price_with_tax', '')) === 0)
				{
					$this->setState('product_price', null);
				}
				else
				{
					// Check if we have an ID or a value
					if ($this->getState('product_tax_id', false))
					{
						// Find the value
						$this->calcsTable->load($this->getState('product_tax_id'));
						$this->setState('product_tax', $this->calcsTable->calc_value);
					}

					$this->setState(
						'product_price',
						$this->getState('price_with_tax') / (1 + ($this->getState('product_tax') / 100))
					);
				}
			}
			elseif (strlen($this->getState('product_price', '')) === 0)
			{
				$this->setState('product_price', null);
			}

			// Check if we need to assign a shopper group
			if (null !== $this->shopper_group_name_price)
			{
				if ($this->getState('shopper_group_name_price', '*') === '*')
				{
					$this->setState('virtuemart_shoppergroup_id', 0);
				}
				else
				{
					$this->setState(
						'virtuemart_shoppergroup_id', $this->helper->getCustomerGroupId($this->shopper_group_name_price)
					);
				}
			}

			// Bind the fields to check for an existing price
			$this->productPriceTable->bind($this->state);

			// Check if the price already exists
			if (!$this->productPriceTable->check())
			{
				// Price doesn't exist
				if (!$this->productPriceTable->get('price_quantity_start'))
				{
					$this->productPriceTable->price_quantity_start = 0;
				}

				if (!$this->productPriceTable->get('price_quantity_end'))
				{
					$this->productPriceTable->price_quantity_end = 0;
				}

				if ($this->productPriceTable->get('override', false) === false)
				{
					$this->productPriceTable->override = 0;
				}

				if (!$this->productPriceTable->get('product_price_publish_up'))
				{
					$this->productPriceTable->product_price_publish_up = '0000-00-00 00:00:00';
				}

				if (!$this->productPriceTable->get('product_price_publish_down'))
				{
					$this->productPriceTable->product_price_publish_down = '0000-00-00 00:00:00';
				}

				// Set the create date if the user has not done so and there is no product_price_id
				if (!$this->productPriceTable->get('created_on'))
				{
					$this->productPriceTable->created_on = $this->date->toSql();
					$this->productPriceTable->created_by = $this->userId;
				}
			}

			// Check if we need to change the shopper group name
			if (null !== $this->shopper_group_name_new)
			{
				if ($this->shopper_group_name_new === '*')
				{
					$this->productPriceTable->virtuemart_shoppergroup_id = 0;
				}
				else
				{
					$this->productPriceTable->virtuemart_shoppergroup_id = $this->helper->getCustomerGroupId(
						$this->shopper_group_name_new
					);
				}
			}

			// Re-bind because the values are loaded during the check
			$this->productPriceTable->bind($this->state);

			// Calculate the new price
			$this->productPriceTable->CalculatePrice();

			/**
			 * Store the price
			 * Add some variables if needed
			 * Set the modified date if the user has not done so
			 */
			if (!$this->productPriceTable->get('modified_on'))
			{
				$this->productPriceTable->set('modified_on', $this->date->toSql());
				$this->productPriceTable->set('modified_by', $this->userId);
			}

			// Store the price
			$this->productPriceTable->store();

			// VM 3.8 update has_prices field
			$this->setState('has_prices', 1);
		}
	}

	/**
	 * Manufacturer import.
	 *
	 * Adds or updates a manufacturer and adds a reference to the product.
	 *
	 * @return  bool Returns true on OK | False on failure.
	 *
	 * @since   3.0
	 *
	 * @throws  Exception
	 */
	private function manufacturerImport()
	{
		$this->log->add('Checking manufacturer', false);

		$manufacturerId = $this->getState('manufacturer_id', false);

		if (!$manufacturerId)
		{
			$manufacturerName = $this->getState('manufacturer_name', false);

			if (!$manufacturerName)
			{
				return false;
			}

			$this->manufacturerTable->load(['title' => $manufacturerName]);

			if ($this->manufacturerTable->id)
			{
				$manufacturerId = $this->manufacturerTable->id;
			}
		}

		if ($manufacturerId)
		{
			$this->manufacturerTable->set('id', $manufacturerId);

			if (!$this->manufacturerTable->check() && !$this->manufacturerTable->store())
			{
				return false;
			}

			$this->setState('manufacturer_id', $manufacturerId);

			$this->manufacturerTable->reset();
		}

		return true;
	}

	/**
	 * Process related products.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 *
	 * @throws  Exception
	 */
	private function processRelatedProducts()
	{
		$relatedproducts = explode('|', $this->related_products);

		$query   = 'INSERT IGNORE INTO ';
		$query   .= $this->db->quoteName('#__csvi_related_products') . ' VALUES ';
		$entries = array();

		foreach ($relatedproducts as $relatedproduct)
		{
			$entries[] = '(' . $this->db->quote($this->product_sku) . ', ' . $this->db->quote($relatedproduct) . ')';
		}

		$query .= implode(',', $entries);
		$this->db->setQuery($query)->execute();

		// Remove any existing product relations
		$this->productCustomfieldTable->deleteRelated(
			$this->virtuemart_product_id, $this->virtuemart_vendor_id, $this->helper->getRelatedId('R')
		);
	}

	/**
	 * Execute any processes to finalize the import.
	 *
	 * @param   array  $fields  list of fields used for import
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 */
	public function getPostProcessing($fields = array())
	{
		// Related products
		if (in_array('related_products', $fields))
		{
			$this->postProcessRelatedProducts();
		}
	}

	/**
	 * Post-process related products.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 *
	 * @throws  Exception
	 */
	private function postProcessRelatedProducts()
	{
		// Get the related products
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('p1.virtuemart_product_id', 'virtuemart_product_id'))
			->select($this->db->quoteName('p2.virtuemart_product_id', 'customfield_value'))
			->from($this->db->quoteName('#__csvi_related_products', 'r'))
			->leftJoin(
				$this->db->quoteName('#__virtuemart_products', 'p1')
				. ' ON ' . $this->db->quoteName('r.product_sku') . ' = ' . $this->db->quoteName('p1.product_sku')
			)
			->leftJoin(
				$this->db->quoteName('#__virtuemart_products', 'p2')
				. ' ON ' . $this->db->quoteName('r.related_sku') . ' = ' . $this->db->quoteName('p2.product_sku')
			);
		$this->db->setQuery($query);
		$relations = $this->db->loadObjectList();
		$this->log->add('Process related products');

		if (!empty($relations))
		{
			// Store the new relations
			foreach ($relations as $related)
			{
				// Build the object to store
				$related->virtuemart_custom_id = $this->helper->getRelatedId('R');
				$related->published            = 0;
				$related->created_on           = $this->date->toSql();
				$related->created_by           = $this->userId;
				$related->modified_on          = $this->date->toSql();
				$related->modified_by          = $this->userId;
				$related->customfield_params   = '';

				// Bind the data
				$this->productCustomfieldTable->bind($related);

				// Store the data
				$this->productCustomfieldTable->store();

				// Clean the table object for next insert
				$this->productCustomfieldTable->reset();
			}

			// Empty the relations table
			$this->db->truncateTable('#__csvi_related_products');
		}
		else
		{
			$this->log->addStats('incorrect', 'COM_CSVI_NO_RELATED_PRODUCTS_FOUND');
		}
	}

	/**
	 * Load the product related tables.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 *
	 * @throws  \CsviException
	 * @throws  \RuntimeException
	 */
	public function loadTables()
	{
		// Load the main tables
		$this->productTable = $this->getTable('Product');
//		$this->mediaTable = $this->getTable('Media');
		$this->productImageTable = $this->getTable('ProductImage');
//		$this->productPriceTable = $this->getTable('ProductPrice');
		$this->taxTable = $this->getTable('Tax');
//		$this->productCustomfieldTable = $this->getTable('ProductCustomfield');
		$this->manufacturerTable = $this->getTable('Manufacturer');
//		$this->productManufacturerTable = $this->getTable('ProductManufacturer');
//		$this->productShoppergroupTable = $this->getTable('ProductShoppergroup');
	}

	/**
	 * Cleaning the product related tables.
	 *
	 * @return  void.
	 *
	 * @since   3.0
	 */
	public function clearTables()
	{
		$this->productTable->reset();
//		$this->mediaTable->reset();
//		$this->productMediaTable->reset();
//		$this->productPriceTable->reset();
//		$this->calcsTable->reset();
//		$this->productCustomfieldTable->reset();
//		$this->manufacturerTable->reset();
//		$this->productManufacturerTable->reset();
//		$this->productShoppergroupTable->reset();
	}

	/**
	 * Get image ID of given image name
	 *
	 * @param   string  $imageName  Image name
	 *
	 * @return  integer ID of the image
	 *
	 * @since   7.18.0
	 */
	private function getMediaId(string $imageName): ?int
	{
		$query = $this->db->getQuery(true)
			->select($this->db->quoteName('virtuemart_media_id'))
			->from($this->db->quoteName('#__virtuemart_medias'))
			->where($this->db->quoteName('file_url') . ' = ' . $this->db->quote($imageName));
		$this->db->setQuery($query);
		$value = $this->db->loadResult();
		$this->log->add(($value) ? 'Found file_url value: ' . $value : 'No file_url value found');

		return $value;
	}
}
