<?php
/**
 * @package     CSVI
 * @subpackage  VirtueMart
 *
 * @author      RolandD Cyber Produksi <contact@rolandd.com>
 * @copyright   Copyright (C) 2006 - @year@ RolandD Cyber Produksi. All rights reserved.
 * @license     GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 * @link        https://rolandd.com
 */

defined('_JEXEC') or die;

use Joomla\CMS\Factory;
use Joomla\CMS\Language\Language;

/**
 * Phoca Cart Product table.
 *
 * @package     CSVI
 * @subpackage  Phocacart
 * @since       8.0.0
 */
class PhocacartTableProduct extends CsviTableDefault
{
	/**
	 * Table constructor.
	 *
	 * @param   string     $table   Name of the database table to model.
	 * @param   string     $key     Name of the primary key field in the table.
	 * @param   JDatabase  &$db     Database driver
	 * @param   array      $config  The configuration parameters array
	 *
	 * @since   8.0.0
	 */
	public function __construct($table, $key, &$db, $config = array())
	{
		parent::__construct('#__phocacart_products', 'id', $db, $config);
	}

	/**
	 * Reset the primary key.
	 *
	 * @return  bool  Returns true on success.
	 *
	 * @since   8.0.0
	 */
	public function reset()
	{
		parent::reset();

		$this->set('id', null);

		return true;
	}

	/**
	 * Create a alias.
	 *
	 * @return  void.
	 *
	 * @since   8.8.0
	 */
	public function checkAlias(): void
	{
		// Is there a product name
		if (!$this->title)
		{
			$jdate = Factory::getDate();
			$this->alias = $jdate->format("Y-m-d-h-i-s") . mt_rand();
		}
		else
		{
			// Create the slug
			$this->alias = $this->createAlias($this->title);

			// Check if the slug exists
			$this->verifyAlias();
		}
	}

	/**
	 * Create a alias.
	 *
	 * @param   string  $name  The string to turn into a alias
	 *
	 * @return  string  The alias created.
	 *
	 * @since   8.8.0
	 */
	public function createAlias(string $name): string
	{
		// Get the correct language
		$language = $this->template->get('language', '', null, 0, false);

		// Transliterate
		$lang = new Language($language);
		$str = $lang->transliterate($name);

		// Trim white spaces at beginning and end of alias and make lowercase
		$str = trim(\Joomla\String\StringHelper::strtolower($str));

		// Remove any duplicate whitespace, and ensure all characters are alphanumeric
		$str = preg_replace('/(\s|[^A-Za-z0-9\-])+/', '-', $str);

		// Trim dashes at beginning and end of alias
		$str = trim($str, '-');

		// If we are left with an empty string, make a date with random number
		if (trim(str_replace('-', '', $str)) == '')
		{
			$jdate = Factory::getDate();
			$str = $jdate->format("Y-m-d-h-i-s") . mt_rand();
		}

		return $str;
	}

	/**
	 * Check if a alias exists.
	 *
	 * @return  void.
	 *
	 * @since   8.8.0
	 */
	private function verifyAlias(): void
	{
		// Check if the slug exists
		$query = $this->db->getQuery(true)
			->select('COUNT(' . $this->db->quoteName($this->_tbl_key) . ')')
			->from($this->_tbl)
			->where($this->db->quoteName('alias') . ' = ' . $this->db->quote($this->alias))
			->where($this->db->quoteName($this->_tbl_key) . ' != ' . (int) $this->id);
		$this->db->setQuery($query);
		$alias = $this->db->loadResult();
		$this->log->add('Check if the product slug exists');

		if ($alias > 0)
		{
			if ($this->get('title'))
			{
				// See how many products we have with the same name
				$query->clear()
					->select(
						'ABS(REPLACE(' . $this->db->quoteName('alias') . ', ' . $this->db->quote($this->alias) . ', ' . $this->db->quote('') . ')) + 1'
					)
					->from($this->_tbl)
					->where(
						'('
						. $this->db->quoteName('alias') . ' = ' . $this->db->quote($this->alias)
						. ' OR '
						. $this->db->quoteName('alias') . ' LIKE ' . $this->db->quote($this->alias . '-_%')
						. ' AND RIGHT(' . $this->db->quoteName('alias') . ', 1) IN (0,1,2,3,4,5,6,7,8,9)'
						. ')'
					)
					->where($this->db->quoteName($this->_tbl_key) . ' != ' . (int) $this->get('id'))
					->order('LENGTH(' . $this->db->quoteName('alias') . '), ' . $this->db->quoteName('alias') . ' ASC');
				$this->db->setQuery($query);
				$items = $this->db->loadColumn();
				$this->log->add('Check how many other products exist with the same slug');

				if ($items)
				{
					$count = array_pop($items);

					$this->alias .= '-' . $count;
				}
			}
			else
			{
				$jdate = Factory::getDate();
				$this->alias .= $jdate->format("Y-m-d-h-i-s") . mt_rand();
			}
		}
	}
}
